﻿using System;
using System.ComponentModel;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Data;
using System.Collections;

using CMS.Controls;
using CMS.Helpers;
using CMS.IO;
using CMS.Base;

namespace CMS
{
    /// <summary>
    /// Basic Bing maps control that automates databinding.
    /// </summary>
    [DefaultProperty("Text"), ToolboxData("<{0}:BasicBingV8Maps runat=server></{0}:BasicBingV8Maps>")]
    public class BasicBingV8Maps : CMSAbstractMap
    {
        #region "Bing service response status constants"

        /// <summary>
        /// Indicates that there is a problem with the service right now.
        /// </summary>
        public const string SERVICE_UNAVAILABLE = "503";

        /// <summary>
        /// Indicates that request could not be completed because there was a problem with the service.
        /// </summary>
        public const string INTERNAL_SERVER_ERROR = "500";

        /// <summary>
        /// Indicates that the request is successful.
        /// </summary>
        public const string OK = "200";

        #endregion


        #region "Private variables"

        private static BasicBingV8Maps mProviderObject = null;
        private static Regex mRegJsonStatus = null;
        private static Regex mRegJsonCoordinates = null;

        #endregion


        #region "Private properties"

        /// <summary>
        /// Regular expression for parsing Json response to get coordinates.
        /// </summary>
        private static Regex RegJsonCoordinates
        {
            get
            {
                if (mRegJsonCoordinates == null)
                {
                    mRegJsonCoordinates = RegexHelper.GetRegex("coordinates\"[\\w\\W]+?(-?\\d+\\.\\d+)[\\w\\W]+?(-?\\d+\\.\\d+)");
                }
                return mRegJsonCoordinates;
            }
        }


        /// <summary>
        /// Regular expression for parsing Json response to get response status.
        /// </summary>
        private static Regex RegJsonStatus
        {
            get
            {
                if (mRegJsonStatus == null)
                {
                    mRegJsonStatus = RegexHelper.GetRegex("statusCode\":(\\d+)");
                }
                return mRegJsonStatus;
            }
        }


        /// <summary>
        /// Provider object.
        /// </summary>
        protected static BasicBingV8Maps ProviderObject
        {
            get
            {
                if (mProviderObject == null)
                {
                    mProviderObject = new BasicBingV8Maps();
                }
                return mProviderObject;
            }
            set
            {
                mProviderObject = value;
            }
        }

        #endregion


        #region "Public methods"

        /// <summary>
        /// Registers necessary Bing javascript files.
        /// </summary>
        /// <param name="page">Page</param>
        /// <param name="mapId">Map Id</param>
        /// <param name="mainScriptPath">Map script path</param>
        public static void RegisterMapScripts(Page page, string mapId, string mainScriptPath)
        {
            string bingScriptUrl = "//www.bing.com/api/maps/mapcontrol";

            // Register Bing API
            ScriptHelper.RegisterClientScriptBlock(page, typeof(string), "bingMapsScriptFile", ScriptHelper.GetScriptTag(bingScriptUrl));

            // Register OnLoad script
            RegisterOnLoadScript(page);

            // Register startup script
            if (!string.IsNullOrEmpty(mapId))
            {
                ScriptHelper.RegisterStartupScript(page, typeof(string), "bingMapsStartup_" + mapId, ScriptHelper.GetScript("addLoadEvent(MapLoad_" + mapId + ");"));
            }

            // Register Microsoft.Maps object
            ScriptHelper.RegisterClientScriptBlock(page, typeof(string), "bingMapsObject", ScriptHelper.GetScript("var MM = Microsoft.Maps;"));

            // Register BingMaps helper javascript file
            RegisterMapFileScript(page, mainScriptPath);
        }


        /// <summary>
        /// Generates static map.
        /// </summary>
        /// <param name="mp">Map properties</param>
        /// <param name="cacheMinutes">Cache minutes</param>
        /// <param name="cacheItemName">Cache item name</param>
        /// <param name="InstanceGUID">Web part instance guid</param>
        public static string GenerateMap(CMSMapProperties mp, int cacheMinutes, string cacheItemName, Guid InstanceGUID)
        {
            double[] geoResponse = null;

            // Cache marker coordinates 
            if (mp.Latitude == null && mp.Longitude == null && !string.IsNullOrEmpty(mp.Location.Trim()) && mp.EnableServerProcessing)
            {
                string viewMode = GetViewMode();

                if (!string.IsNullOrEmpty(cacheItemName))
                {
                    cacheItemName += "|" + viewMode;
                }

                // Custom cache key name for LiveSite and the other view modes 
                string customCacheItemName = CacheHelper.GetCacheItemName(cacheItemName, mp.Location, "StaticBingMap", viewMode);
                using (var cs = new CachedSection<double[]>(ref geoResponse, cacheMinutes, true, null, customCacheItemName, null))
                {
                    if (cs.LoadData)
                    {
                        // Cache response
                        geoResponse = GetGeoCoordinates(mp.Location, mp.MapKey);

                        // Flush cache if web part properties are changed
                        if (cs.Cached)
                        {
                            cs.CacheDependency = CacheHelper.GetCacheDependency("webpartinstance|" + InstanceGUID.ToString().ToLowerCSafe());
                        }

                        cs.Data = geoResponse;
                    }
                }
            }

            if (geoResponse != null)
            {
                // Get coordinates from cache
                mp.Latitude = geoResponse[0];
                mp.Longitude = geoResponse[1];
            }

            return ProviderObject.GenerateMapInternal(mp, null, null);
        }

        #endregion


        #region "Private methods"

        /// <summary>
        /// OnItemDataBound override.
        /// </summary>
        protected override void OnItemDataBound(RepeaterItemEventArgs e)
        {
            base.OnItemDataBound(e);
            // Add each content item to javascript array
            e.Item.Controls.AddAt(0, new LiteralControl("content_" + MapProperties.MapId + ".push({'cont':'##TR##"));
            e.Item.Controls.Add(new LiteralControl("##TRE##'});\n"));
        }


        /// <summary>
        /// OnPreRender override.
        /// </summary>
        protected override void OnPreRender(EventArgs e)
        {
            base.OnPreRender(e);

            if (!DataHelper.DataSourceIsEmpty(DataSource) || !HideControlForZeroRows || !StopProcessing)
            {
                // Register Bing javascript files
                RegisterMapScripts(Page, MapProperties.MapId, MainScriptPath);
            }
        }


        /// <summary>
        /// Render override.
        /// </summary>
        protected override void Render(HtmlTextWriter writer)
        {
            // Is empty, apply the hiding rules
            if (DataHelper.DataSourceIsEmpty(DataSource))
            {
                // Hide the control when HideControlForZeroRows is set
                if (HideControlForZeroRows)
                {
                    return;
                }
                else if (!string.IsNullOrEmpty(ZeroRowsText))
                {
                    // If the ZeroRowsText is set, display it instead if the control itself
                    writer.Write(ZeroRowsText);
                    return;
                }
            }

            StringBuilder markerArray = new StringBuilder();
            StringWriter sw = new StringWriter(markerArray);
            Html32TextWriter mwriter = new Html32TextWriter(sw);
            base.Render(mwriter);
            sw.Dispose();

            ArrayList contentArray = new ArrayList();

            // Cache coordinates if server processing is enabled
            if (MapProperties.EnableServerProcessing)
            {
                string viewMode = GetViewMode();

                if (!string.IsNullOrEmpty(CacheItemName))
                {
                    CacheItemName += "|" + viewMode;
                }

                // Custom cache key name for LiveSite and the other view modes 
                string customCacheItemName = CacheHelper.GetCacheItemName(CacheItemName, MapProperties.MapName, "BingMap", viewMode);
                using (var cs = new CachedSection<ArrayList>(ref contentArray, CacheMinutes, true, null, customCacheItemName, null))
                {
                    if (cs.LoadData)
                    {
                        contentArray = GetContentArray(DataSource);

                        if (cs.Cached)
                        {
                            cs.CacheDependency = CacheHelper.GetCacheDependency(CacheDependencies);
                        }

                        cs.Data = contentArray;
                    }
                }
            }
            else
            {
                contentArray = GetContentArray(DataSource);
            }

            // Ensure Page presence for client geocoding
            if (MapProperties.Page == null)
            {
                MapProperties.Page = Page;
            }

            // Generate map with corrected marker content array
            outputHtml += GenerateMapInternal(MapProperties, RepairInputString(markerArray.ToString()), contentArray);
            // Render Html
            writer.Write(outputHtml);
        }


        /// <summary>
        /// Returns ArrayList of content table from DataSource.
        /// </summary>
        /// <param name="dataSource">DataSource object</param>
        private ArrayList GetContentArray(object dataSource)
        {
            ArrayList contentArray = new ArrayList();

            string latitudeField = MapProperties.LatitudeField;
            string longitudeField = MapProperties.LongitudeField;
            string tooltipField = MapProperties.ToolTipField;
            string locationField = MapProperties.LocationField;
            string iconField = MapProperties.IconField;
            string mapKey = MapProperties.MapKey;

            if (!DataHelper.DataSourceIsEmpty(dataSource))
            {
                DataTable dt = DataHelper.GetDataTable(dataSource);

                foreach (DataRow dr in dt.Rows)
                {
                    double? lat = null;
                    // Ensure latitude
                    string latTemp = DataHelper.GetNotEmpty(DataHelper.GetDataRowValue(dr, latitudeField), "");
                    if (!string.IsNullOrEmpty(latTemp))
                    {
                        lat = ValidationHelper.GetDouble(latTemp, 0);
                    }
                    double? lng = null;
                    // Ensure longitude    
                    string lngTemp = DataHelper.GetNotEmpty(DataHelper.GetDataRowValue(dr, longitudeField), "");
                    if (!string.IsNullOrEmpty(lngTemp))
                    {
                        lng = ValidationHelper.GetDouble(lngTemp, 0);
                    }
                    string tooltip = DataHelper.GetStringValue(dr, tooltipField).Trim();

                    string location = String.Empty;
                    if (!string.IsNullOrEmpty(locationField))
                    {
                        foreach (string loc in locationField.Split(';'))
                        {
                            location += DataHelper.GetStringValue(dr, loc).Trim() + ",";
                        }

                        location = location.Trim(',');
                    }

                    string iconURL = DataHelper.GetStringValue(dr, iconField).Trim();

                    // Ensure correct coordinates
                    lat = EnsureLatitude(lat);
                    lng = EnsureLongitude(lng);

                    if (lat == null && lng == null && MapProperties.EnableServerProcessing && !string.IsNullOrEmpty(location))
                    {
                        // Get coordinates from Bing service
                        double[] geoResult = GetGeoCoordinates(location, mapKey);
                        lat = geoResult[0];
                        lng = geoResult[1];
                    }
                    contentArray.Add(new ArrayList() { lat, lng, location, tooltip, iconURL });
                }
            }

            return contentArray;
        }


        /// <summary>
        /// Generates complete map code.
        /// </summary>
        /// <param name="mp">Map properties</param>
        /// <param name="markerArray">Marker array</param>
        /// <param name="contentArray">ArrayList of content table</param>
        private string GenerateMapInternal(CMSMapProperties mp, string markerArray, ArrayList contentArray)
        {
            StringBuilder result = new StringBuilder();

            // Initialize local variables from CMSMapProperties
            string mapId = mp.MapId;
            string mapName = mp.MapName;
            int scale = mp.Scale;
            string height = mp.Height;
            string width = mp.Width;
            string location = mp.Location;
            string mapType = mp.MapType;
            string mapKey = mp.MapKey;

            // Ensure correct map type
            if (string.IsNullOrEmpty(mapType))
            {
                mapType = "road";
            }

            // Marker content - global variable
            result.AppendFormat("var content_{0} = [];\nvar {1} = null;\n", mapId, mapName);

            // Check if single marker is generated - detail mode
            bool detailMode = false;

            double? lat;
            double? lng;

            // Ensure correct latitude and longitude
            if (contentArray != null)
            {
                // Add marker array
                result.Append(markerArray);

                if (contentArray.Count == 1)
                {
                    detailMode = true;

                    ArrayList coordinates = contentArray[0] as ArrayList;
                    // Latitude                    
                    lat = (double?)coordinates[0];
                    // Longitude
                    lng = (double?)coordinates[1];
                }
                else
                {
                    lat = mp.Latitude;
                    lng = mp.Longitude;
                }
            }
            else
            {
                detailMode = true;
                result.AppendFormat("content_{0}.push({{'cont': '{1}'}});\n", mapId, RepairMarkerContent(mp.Content));

                lat = mp.Latitude;
                lng = mp.Longitude;
            }

            lat = EnsureLatitude(lat);
            lng = EnsureLongitude(lng);

            // Register map geo code handler for map center
            if (!mp.EnableServerProcessing && !string.IsNullOrEmpty(mp.Location.Trim()) && lat == null && lng == null && !detailMode)
            {
                result.AppendFormat(@"
                 function callback_{0}(res) 
                 {{
                    if (res && res.resourceSets && res.resourceSets.length > 0 && res.resourceSets[0].resources && res.resourceSets[0].resources.length > 0) {{
                       var loc = new MM.Location(res.resourceSets[0].resources[0].point.coordinates[0], res.resourceSets[0].resources[0].point.coordinates[1]);
                       {0}.setView({{ zoom: {1}, center: new MM.Location(loc.latitude, loc.longitude) }});
                    }}
                    else {{
                       {0}.setView({{ zoom: {1}, center: new MM.Location(0, 0) }});
                    }} 
                 }}", mp.MapName, scale);
            }

            result.AppendFormat(@"
function infBox_{0}(e){{
 showInfo({1}, e.target, {2});
}}", mapId, mapName, mp.ZoomScale);

            result.AppendFormat(@"
function MapLoad_{1}(){{
if ({0} != null){{
    {0} = null;
}}
{0} = new MM.Map('#{0}'
,
{{
credentials: '{2}',
enableSearchLogo: enableSearchLogo,
mapTypeId: MM.MapTypeId.{5},
width: {3},
height: {4},
showDashboard: " + (mp.ShowNavigationControl ? "true,\n" : "false,\n") + @"
disablePanning: " + (mp.EnableMapDragging ? "false,\n" : "true,\n") + @"
showScalebar: " + (mp.ShowScaleControl ? "true" : "false"), mapName, mapId, HTMLHelper.HTMLEncode(mapKey), width, height, mapType);

            bool clientProcMapFocus = false;
            // Center of the map
            if (!detailMode)
            {
                // Map center according to default lat and lng
                if (lat != null || lng != null || (string.IsNullOrEmpty(location.Trim())))
                {
                    // If latitude or longitude is not specified, set to zero
                    lat = lat ?? 0;
                    lng = lng ?? 0;

                    result.AppendFormat(",\nzoom: {0},\n center: new MM.Location({1}, {2})\n", scale, lat.ToString().Replace(",", "."), lng.ToString().Replace(",", "."));
                }
                else if (mp.EnableServerProcessing)
                {
                    // Get coordinates from Bing service for map center
                    double[] latLon = GetGeoCoordinates(location.Trim(), mp.MapKey.Trim());
                    // Server processing according to location
                    result.AppendFormat(",\nzoom: {0},\n center: new MM.Location({1}, {2})\n", scale, latLon[0], latLon[1]);
                }
                else
                {
                    clientProcMapFocus = true;
                }
            }
            result.Append("\n});\n");

            if (clientProcMapFocus)
            {
                // Client processing according to location for default map focus
                result.AppendFormat(@"
function bingRequest_{1}(credentials) {{
    var url = ""//dev.virtualearth.net/REST/v1/Locations/"" + {2} + ""?output=json&jsonp=callback_{0}&key="" + credentials;
    callBingService(url);
}}                    
{0}.getCredentials(bingRequest_{1});
                    ", mp.MapName, mp.MapId, ScriptHelper.GetString(location.Trim()));
            }

            // Generate markers from ArrayList
            if (contentArray != null)
            {
                // Marker array index
                int index = 0;
                foreach (ArrayList item in contentArray)
                {
                    //item => [0] = latitude, [1] = longitude, [2] = location, [3] = tooltip, [4] = iconURL
                    double? iLatitude = (double?)item[0];
                    double? iLongitude = (double?)item[1];
                    string iLocation = ValidationHelper.GetString(item[2], "");
                    string iTooltip = ValidationHelper.GetString(item[3], "");
                    string iIconURL = ValidationHelper.GetString(item[4], "");

                    // Generate marker
                    result.Append(GenerateMapMarker(iLatitude, iLongitude, iLocation, iTooltip, index, mp, detailMode, iIconURL));
                    index++;
                }
            }
            else
            {
                // Generate single marker with same map center
                result.Append(GenerateMapMarker(lat, lng, location, mp.ToolTip, 0, mp, true, mp.IconURL));
            }

            //End of main map function
            result.Append("}");

            // Ensure width height definition
            if (ValidationHelper.IsInteger(height))
            {
                height = height + "px";
            }
            if (ValidationHelper.IsInteger(width))
            {
                width = width + "px";
            }

            StringBuilder mapHolder = new StringBuilder();

            // Add map holder
            mapHolder.AppendFormat("<div style=\"position:relative; width:{0}; height:{1};\" id=\"{2}\"></div>", width, height, mapName);

            // Return complete map code with javascript tags and add map holder to result
            return ScriptHelper.GetScript(result.ToString()) + mapHolder;
        }


        /// <summary>
        /// Generates Bing map marker.
        /// </summary>
        /// <param name="lat">Marker latitude</param>
        /// <param name="lng">Marker longitude</param>
        /// <param name="location">Marker location</param>
        /// <param name="toolTip">Marker tooltip</param>
        /// <param name="markerIndex">Marker array index</param>
        /// <param name="mp">Map Properties</param>
        /// <param name="setMapCenter">Indicates if map center is same as single marker</param>
        /// <param name="iconURL">Custom marker URL</param>
        private string GenerateMapMarker(double? lat, double? lng, string location, string toolTip, int markerIndex, CMSMapProperties mp, bool setMapCenter, string iconURL)
        {
            StringBuilder result = new StringBuilder();

            // Marker content array reference
            string markerRef = "content_" + mp.MapId + "[" + markerIndex + "]['cont']";

            // Custom marker icon
            if (!string.IsNullOrEmpty(iconURL))
            {
                iconURL = ResolveUrl(iconURL);
            }

            if ((lat == null) && (lng == null) && !mp.EnableServerProcessing && !string.IsNullOrEmpty(location.Trim()))
            {
                // Get coordinates on client side and add marker to map
                result.AppendFormat(@"
                {0}.getCredentials(bingRequest_{1}_{2});
                function bingRequest_{1}_{2}(credentials)
                {{
                var url = ""//dev.virtualearth.net/REST/v1/Locations/"" + {3} + ""?output=json&jsonp=callback_{0}_{2}&key="" + credentials;
                callBingService(url);
                }}", mp.MapName, mp.MapId, markerIndex, ScriptHelper.GetString(location.Trim()));

                StringBuilder geoMarkerScript = new StringBuilder();
                geoMarkerScript.AppendFormat(@"
                 function callback_{0}_{2}(res) 
                 {{
                    if (res && res.resourceSets && res.resourceSets.length > 0 && res.resourceSets[0].resources && res.resourceSets[0].resources.length > 0) {{
                       var loc = new MM.Location(res.resourceSets[0].resources[0].point.coordinates[0], res.resourceSets[0].resources[0].point.coordinates[1]);
                       " + (setMapCenter ? "{0}.setView({{ zoom: {4}, center: new MM.Location(loc.latitude, loc.longitude) }});\n" : "") + @"
                       MM.Events.addHandler(addBingMarker({0}, loc.latitude, loc.longitude, {5}, {6}, {7}), 'click', infBox_{8});
                    }}
                 }}", mp.MapName, mp.MapId, markerIndex, ScriptHelper.GetString(location.Trim()), mp.Scale, ScriptHelper.GetString(toolTip.Trim()), markerRef.Trim(), ScriptHelper.GetString(iconURL), mp.MapId);
                ScriptHelper.RegisterStartupScript(mp.Page, typeof(string), "bingGeoMarkerScript_" + mp.MapName + "_" + markerIndex, ScriptHelper.GetScript(geoMarkerScript.ToString()));
            }
            else if ((lat != null) || (lng != null))
            {
                // If latitude or longitude is not specified, set to zero
                lat = lat ?? 0;
                lng = lng ?? 0;

                // Add marker to map without server/client geo processing
                result.AppendFormat((setMapCenter ? "{0}.setView({{ zoom: {3}, center: new MM.Location({1}, {2}) }});\n" : "") +
                                    @"MM.Events.addHandler(addBingMarker({0}, {1}, {2}, {4}, {5}, {6}), 'click', infBox_{7});
                       ", mp.MapName, lat.ToString().Replace(",", "."), lng.ToString().Replace(",", "."), mp.Scale, ScriptHelper.GetString(toolTip.Trim()), markerRef.Trim(), ScriptHelper.GetString(iconURL), mp.MapId);
            }
            return result.ToString();
        }


        /// <summary>
        /// Returns coordinates according to location address from Bing service.
        /// </summary>
        /// <param name="location">Location</param>
        /// <param name="mapKey">Bing map key</param>
        private static double[] GetGeoCoordinates(string location, string mapKey)
        {
            double[] result = { 0, 0 };

            try
            {
                bool repeatRequest;
                int interval = 150;
                int trials = 0;

                // Create WebClient object
                WebClient client = new WebClient();

                do
                {
                    client.Headers.Add("Content-Type", "application/x-www-form-urlencoded");
                    // Create url
                    string url = URLHelper.AddParameterToUrl("https://dev.virtualearth.net/REST/v1/Locations", "query", HttpUtility.UrlEncode(location));
                    url = URLHelper.AddParameterToUrl(url, "key", mapKey);

                    // JSON response from bing service
                    string jsonResponse = client.DownloadString(url);
                    client.Dispose();

                    // Get response status according to regular expression
                    Match matchStatus = RegJsonStatus.Match(jsonResponse);

                    if (matchStatus.Success)
                    {
                        string statusCode = matchStatus.Groups[1].Value;

                        if (statusCode.EqualsCSafe(OK, true))
                        {
                            // Get coordinates from response according to regular expression
                            Match mtchCoordinates = RegJsonCoordinates.Match(jsonResponse);
                            if (mtchCoordinates.Success)
                            {
                                result[0] = ValidationHelper.GetDouble(mtchCoordinates.Groups[1].Value, 0);
                                result[1] = ValidationHelper.GetDouble(mtchCoordinates.Groups[2].Value, 0);
                            }
                            repeatRequest = false;
                        }
                        else if (statusCode.EqualsCSafe(SERVICE_UNAVAILABLE, true) || statusCode.EqualsCSafe(INTERNAL_SERVER_ERROR, true))
                        {
                            repeatRequest = true;
                            // Delay next request
                            Thread.Sleep(interval);
                            // Increase interval
                            interval *= 2;
                            trials++;
                        }
                        else
                        {
                            repeatRequest = false;
                        }
                    }
                    else
                    {
                        repeatRequest = false;
                    }
                } while (repeatRequest && trials < 4);
            }
            catch
            {
            }

            return result;
        }

        #endregion
    }
}